import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spl
norm = np.linalg.norm
import matplotlib.pyplot as plt

import pdb

# min c.dot(x) : x >= 0, Mx <= b
d = 3
n = 4
M = sp.csc_matrix(np.array([[2,4,5,7], [1,1,2,2], [1,2,3,3]]), dtype='float')
c = -np.array([7,9,18,17])
b = np.array([41,17,24])


gamma = 0.9
u, s, v = spl.svds(M,1)
NM = s[0] ** 2
sigma = 1 * np.sqrt(gamma) / np.sqrt(NM)
tau = gamma / (sigma * NM)

def L(x, y):
    if (x < 0).any():
        return np.inf
    if (y < 0).any():
        return np.inf
    return c.dot(x) + y.dot(M @ x - b)

def proj_Ystar(y):
    # Ystar = {[y1, 3-y1, 4-y1], y1 \in [0,3]}
    y0_star = np.array([0, 3, 4])
    y1_star = np.array([3, 0, 1])
    ll = -(y - y0_star).dot(y0_star - y1_star) / norm(y0_star - y1_star)**2
    ll = max(0, min(1, ll))
    return ll * y1_star + (1-ll) * y0_star


def smoothed_gap(x, y, beta, x_star=None, y_star=None):
    if x_star == None:
        x_star = np.array([3, 0, 7, 0])
    if y_star == None:
        y_star = proj_Ystar(y)
    betax = beta[0]
    betay = beta[1]

    sm_gap_x = c.dot(x) + betay/2 * norm(np.maximum(0, y_star+1/betay*(M.dot(x)-b)))**2 - betay/2 * norm(y_star)**2 - c.dot(x_star)
    sm_gap_y = b.dot(y) + betax/2 * norm(np.maximum(0, x_star-1/betax*(M.T.dot(y) + c)))**2 - betax/2 * norm(x_star)**2 - b.dot(y_star)
    sm_gap = sm_gap_x + sm_gap_y

    if sm_gap < -1e-12:
        pdb.set_trace()
    else:
        sm_gap = max(0, sm_gap)
    
    return sm_gap

def grad_lag(x,y, tau, sigma):
    x_star = np.array([3, 0, 7, 0])
    y_star = proj_Ystar(y)

    gradient = np.zeros(len(x)+len(y))
    for i in range(len(x)):
        if x[i] > 0:
            gradient[i] = c[i] + M.T[i].dot(y)
        if x[i] == 0:
            gradient[i] = np.minimum(0, c[i] + M.T[i].dot(y))
    n = len(x)
    for j in range(len(y)):
        if y[j] > 0:
            gradient[n+j] = M[j].dot(x) - b[j]
        if y[j] == 0:
            gradient[n+j] = np.minimum(0, M[j].dot(x) - b[j])
    nn = np.sqrt(norm(gradient[0:len(x)])**2*tau+norm(gradient[len(x):])**2*sigma)
    return [nn, nn/np.sqrt(norm(x - x_star)**2/tau + norm(y - y_star)**2/sigma)]


def pdhg(x_init, y_init, tau, sigma, K, betas=[], adaptive_restart=0):
    x = x_init.copy()
    y = y_init.copy()
    avgx = x.copy()
    avgy = y.copy()
    x_star = np.array([3, 0, 7, 0])
    y_star = np.array([2.75323656, 0.24676344, 1.24676344])


    Nx = np.zeros(K)
    Ny = np.zeros(K)
    tV = np.zeros((K,2))
    Gap = np.zeros((K,3+len(betas)*2))
    Grad_lag = np.zeros((K,2))

    for k in range(K):
        x_ = np.maximum(0, x - tau * (M.T @ y + c))
        y_ = np.maximum(0, y + sigma * (M @ x_ - b))
        tV[k,0] = -tau/2. * norm(M.T @ (y_ - y))**2 + 0.5 * norm(x_ - x)**2/tau\
                + 0.5 * norm(y_ - y)**2/sigma
        tV[k,1] = L(x_, y_star) - L (x_star, y_)
        
        x = x_ - tau * M.T @ (y_ - y)
        y = y_.copy()
        avgx = k/(k+1) * avgx + 1./(k+1)*x_
        avgy = k/(k+1) * avgy + 1./(k+1)*y_
    
        Nx[k] = norm(x - x_star)
        Ny[k] = norm(y - proj_Ystar(y))
        Nx_ = norm(x_ - x_star)
        Ny_ = norm(y_ - proj_Ystar(y_))
        V = 0.5/tau * Nx_**2 + 0.5/sigma*Ny_**2
        Gap[k,0] = c.dot(x_) + b.dot(y_)
        for i_beta, beta in enumerate(betas):
            sm_gap = smoothed_gap(x_, y_, [beta/tau, beta/sigma])
            Gap[k,1+i_beta*2:3+i_beta*2] = [sm_gap, sm_gap/V]
        sm_gap = smoothed_gap(avgx, avgy, [2/(k+1)/tau, 2/(k+1)/sigma])
        Gap[k, -2:] = [sm_gap, sm_gap/V]
        Grad_lag[k,:] = grad_lag(x,y, tau, sigma)

        if adaptive_restart == 1 and k > 9 and Gap[k,-2] < 0.25 * smoothed_gap(x_init, y_init, [2/(k+1)/tau, 2/(k+1)/sigma]):
            print(k)
            Nx = Nx[:k+1]
            Ny = Ny[:k+1]
            tV = tV[:k+1]
            Gap = Gap[:k+1,:]
            break

    return x, y, avgx, avgy, Nx, Ny, tV, Gap, Grad_lag


def rapdhg(x_init, y_init, tau, sigma, period, K, betas=[], adaptive_restart=0):
    x = x_init.copy()
    y = y_init.copy()

    Nx = np.zeros(K)
    Ny = np.zeros(K)
    Gap = np.zeros((K, 3+len(betas)*2))
    Grad_lag = np.zeros((K, 2))
    ktot = 0
    j = 0

    while ktot < K:
        x_in, y_in, avgx, avgy, Nx_in, Ny_in, tV, Gap_in, Grad_lag_in = pdhg(x, y,  tau, sigma,
                                                                min(period, K-ktot), betas, adaptive_restart)
        x = avgx
        y = avgy
        # x = x - tau * period * (A.T @ avgy + c)  # stable mais lent
        # y = y + sigma * period * (A @ avgx - b)
        Nx[ktot:ktot+len(Nx_in)] = Nx_in
        Ny[ktot:ktot+len(Ny_in)] = Ny_in
        Gap[ktot:ktot+Gap_in.shape[0],:] = Gap_in
        Grad_lag[ktot:ktot+Grad_lag_in.shape[0],:] = Grad_lag_in
        j = j + 1
        ktot = ktot + len(Nx_in)

    return x, y, x_in, y_in, Nx, Ny, Gap, Grad_lag


def grad_smoothed_ineq(x, y, sigma):
    return M.T.dot(np.maximum(0, y + 1/sigma*(M.dot(x) - b)))

def inner_ppa_restarted_fista(x_init, y_init, sigma, K):
    # minimise_x c.dot(x) + sigma/2 ||max(O, y_init + 1/sigma(Mx - b))||**2 - sigma/2 ||y_init||**2 + sigma/2 ||x-x_init||**2 + I_{x>=0}
    x = x_init.copy()
    xe = x_init.copy()
    theta = 1.
    x_star = np.array([3, 0, 7, 0])
    y_star = np.array([2.75323656, 0.24676344, 1.24676344])
    Nx = np.zeros(K)
    Ny = np.zeros(K)
    restart_period = np.ceil(NM/sigma)
    
    for k in range(K):
        xg = (1-theta) * x + theta * xe
        grad = c + grad_smoothed_ineq(xg, y_init, sigma) + sigma*(x-x_init)
        xe = np.maximum(0, xe - sigma/NM**2/theta*grad)
        x = (1-theta) * x + theta * xe
        theta = 0.5 * (np.sqrt(theta**4 + 4*theta**2) - theta**2)
        Nx[k] = norm(x - x_star)
        Ny[k] = norm(y_init - proj_Ystar(y_init))
        if norm(x - xg)/theta < 1e-14:
            Nx = Nx[:k]
            Ny = Ny[:k]
            break
        if k % restart_period == restart_period-1:
            xe = x.copy()
            theta = 1
    print('inner test ppa', norm(x - xg)/theta + (theta==1))
    return x, Nx, Ny

def ppa(x_init, y_init, sigma, period, K):
    x = x_init.copy()
    y = y_init.copy()
    Nx = np.zeros(K)
    Ny = np.zeros(K)
    j = 0
    ktot = 0
    while ktot < K:
        x, Nx_in, Ny_in = inner_ppa_restarted_fista(x, y, sigma, min(period, K-ktot))
        y = np.maximum(0, y + 1/sigma * (M.dot(x) - b))
        Nx[ktot:ktot+len(Nx_in)] = Nx_in
        Ny[ktot:ktot+len(Ny_in)] = Ny_in
        ktot = ktot+len(Nx_in)
        j=j+1
    return x, y, Nx, Ny


K = 3000
period = 200  #int(np.ceil(np.exp(1) * NM / msr)) 

inner_max_iter_ppa = K//10
sigma_ppa = 26
# x3, y3, Nx3, Ny3 = ppa(np.zeros(n), np.zeros(d), sigma_ppa, inner_max_iter_ppa, K)

betas = [1e-3, 1e-2, 1e-1, 1]
x, y, avgx, avgy, Nx, Ny, tV, Gap, Grad_lag = pdhg(np.zeros(n), np.zeros(d), tau, sigma, K, betas)
x2, y2, avgx, avgy, Nx2, Ny2, Gap2, Grad_lag2 = rapdhg(np.zeros(n), np.zeros(d), tau, sigma, period, K, betas, adaptive_restart=0)
V = Nx**2 / (2*tau) + Ny**2 / (2*sigma)

beta = 1/100
eta = 2/200
plt.figure(1)
cmap = plt.get_cmap(name='tab10')
plt.semilogy(np.arange(K), Nx ** 2 / tau + Ny ** 2 / sigma, c=cmap(0))
plt.semilogy(np.arange(K), Nx2 ** 2 / tau + Ny2 ** 2 / sigma, c=cmap(1))
plt.semilogy(np.arange(K), 10. / (1+beta*eta) ** (np.arange(K)), '--', c=cmap(0))
plt.semilogy(np.arange(K), 10. * np.exp(-np.log(2) * min([beta/2, eta/4])) ** (np.arange(K)), '--', c=cmap(1))
             # np.arange(K), Nx3 ** 2 / tau + Ny3 ** 2 / sigma)
             # np.arange(K), V[0] * (1 - (msr/NM)**2) ** np.arange(K),
             # np.arange(K), V[0] * (1 - msr/NM) ** np.arange(K))
plt.xlabel('iteration')
plt.ylabel('squared distance to the set of saddle points')
plt.legend(['PDHG', 'RAPDHG', 'rate by Th. 1', 'rate by Prop. 14'])  # , 'ppa-fista'])
plt.show()

plt.figure(2)
plt.loglog(betas, np.min(Gap[:, 2:-2:2], axis=0))

plt.show()
